import os
import warnings
import xml.etree.ElementTree as ET

import leather.svg as svg
from leather import theme
from leather.axis import Axis
from leather.data_types import Date, DateTime
from leather.scales import Linear, Scale, Temporal
from leather.series import CategorySeries, Series
from leather.shapes import Bars, Columns, Dots, Line
from leather.utils import DIMENSION_NAMES, Box, IPythonSVG, X, Y


class Chart:
    """
    Container for all chart types.

    :param title:
        An optional title that will be rendered at the top of the chart.
    """
    def __init__(self, title=None):
        self._title = title
        self._series_colors = theme.default_series_colors

        self._layers = []
        self._types = [None, None]
        self._scales = [None, None]
        self._axes = [None, None]

    def _palette(self):
        """
        Return a generator for series colors.
        """
        return (color for color in self._series_colors)

    def set_x_scale(self, scale):
        """
        Set the X :class:`.Scale` for this chart.
        """
        self._scales[X] = scale

    def set_y_scale(self, scale):
        """
        See :meth:`.Chart.set_x_scale`.
        """
        self._scales[Y] = scale

    def add_x_scale(self, domain_min, domain_max):
        """
        Create and add a :class:`.Scale`.

        If the provided domain values are :class:`date` or :class:`datetime`
        then a :class:`.Temporal` scale will be created, otherwise it will
        :class:`.Linear`.

        If you want to set a custom scale class use :meth:`.Chart.set_x_scale`
        instead.
        """
        scale_type = Linear

        if isinstance(domain_min, Date.types) or isinstance(domain_min, DateTime.types):
            scale_type = Temporal

        self.set_x_scale(scale_type(domain_min, domain_max))

    def add_y_scale(self, domain_min, domain_max):
        """
        See :meth:`.Chart.add_x_scale`.
        """
        scale_type = Linear

        if isinstance(domain_min, Date.types) or isinstance(domain_min, DateTime.types):
            scale_type = Temporal

        self.set_y_scale(scale_type(domain_min, domain_max))

    def set_x_axis(self, axis):
        """
        Set an :class:`.Axis` class for this chart.
        """
        self._axes[X] = axis

    def set_y_axis(self, axis):
        """
        See :meth:`.Chart.set_x_axis`.
        """
        self._axes[Y] = axis

    def add_x_axis(self, ticks=None, tick_formatter=None, name=None):
        """
        Create and add an X :class:`.Axis`.

        If you want to set a custom axis class use :meth:`.Chart.set_x_axis`
        instead.
        """
        self._axes[X] = Axis(ticks, tick_formatter, name)

    def add_y_axis(self, ticks=None, tick_formatter=None, name=None):
        """
        See :meth:`.Chart.add_x_axis`.
        """
        self._axes[Y] = Axis(ticks, tick_formatter, name)

    def add_series(self, series, shape):
        """
        Add a data :class:`.Series` to the chart. The data types of the new
        series must be consistent with any series that have already been added.

        There are several shortcuts for adding different types of data series.
        See :meth:`.Chart.add_bars`, :meth:`.Chart.add_columns`,
        :meth:`.Chart.add_dots`, and :meth:`.Chart.add_line`.
        """
        if self._layers and isinstance(self._layers[0][0], CategorySeries):
            raise RuntimeError('Additional series can not be added to a chart with a CategorySeries.')

        if isinstance(series, CategorySeries):
            self._types = series._types
        else:
            for dim in [X, Y]:
                if not self._types[dim]:
                    self._types[dim] = series._types[dim]
                elif series._types[dim] is not self._types[dim]:
                    raise TypeError(f'Can\'t mix axis-data types: {series._types[dim]} and {self._types[dim]}')

        shape.validate_series(series)

        self._layers.append((
            series,
            shape
        ))

    def add_bars(self, data, x=None, y=None, name=None, fill_color=None):
        """
        Create and add a :class:`.Series` rendered with :class:`.Bars`.

        Note that when creating bars in this way the order of the series data
        will be reversed so that the first item in the series is displayed
        as the top-most bar in the graphic. If you don't want this to happen
        use :meth:`.Chart.add_series` instead.
        """
        self.add_series(
            Series(list(reversed(data)), x=x, y=y, name=name),
            Bars(fill_color)
        )

    def add_columns(self, data, x=None, y=None, name=None, fill_color=None):
        """
        Create and add a :class:`.Series` rendered with :class:`.Columns`.
        """
        self.add_series(
            Series(data, x=x, y=y, name=name),
            Columns(fill_color)
        )

    def add_dots(self, data, x=None, y=None, name=None, fill_color=None, radius=None):
        """
        Create and add a :class:`.Series` rendered with :class:`.Dots`.
        """
        self.add_series(
            Series(data, x=x, y=y, name=name),
            Dots(fill_color, radius)
        )

    def add_line(self, data, x=None, y=None, name=None, stroke_color=None, width=None, stroke_dasharray=None):
        """
        Create and add a :class:`.Series` rendered with :class:`.Line`.
        """
        self.add_series(
            Series(data, x=x, y=y, name=name),
            Line(stroke_color, width, stroke_dasharray)
        )

    def _validate_dimension(self, dimension):
        """
        Validates that the given scale and axis are valid for the data that
        has been added to this chart. If a scale or axis has not been set,
        generates automated ones.
        """
        scale = self._scales[dimension]
        axis = self._axes[dimension]

        if not scale:
            scale = Scale.infer(self._layers, dimension, self._types[dimension])
        else:
            for series, shape in self._layers:
                if not scale.contains(series.min(dimension)) or not scale.contains(series.max(dimension)):
                    d = DIMENSION_NAMES[dimension]
                    warnings.warn(
                        'Data contains values outside %s scale domain. '
                        'All data points may not be visible on the chart.' % d
                    )

                    # Only display once per axis
                    break

        if not axis:
            axis = Axis()

        return (scale, axis)

    def to_svg_group(self, width=None, height=None):
        """
        Render this chart to an SVG group element.

        This can then be placed inside an :code:`<svg>` tag to make a complete
        SVG graphic.

        See :meth:`.Chart.to_svg` for arguments.
        """
        width = width or theme.default_width
        height = height or theme.default_height

        if not self._layers:
            raise ValueError('You must add at least one series to the chart before rendering.')

        if isinstance(theme.margin, float):
            default_margin = width * theme.margin

            margin = Box(
                top=default_margin,
                right=default_margin,
                bottom=default_margin,
                left=default_margin
            )
        elif isinstance(margin, int):
            margin = Box(margin, margin, margin, margin)
        elif not isinstance(margin, Box):
            margin = Box(*margin)

        # Root / background
        root_group = ET.Element('g')

        root_group.append(ET.Element(
            'rect',
            x=str(0),
            y=str(0),
            width=str(width),
            height=str(height),
            fill=theme.background_color
        ))

        # Margins
        margin_group = ET.Element('g')
        margin_group.set('transform', svg.translate(margin.left, margin.top))

        margin_width = width - (margin.left + margin.right)
        margin_height = height - (margin.top + margin.bottom)

        root_group.append(margin_group)

        # Header
        header_group = ET.Element('g')

        header_margin = 0

        if self._title:
            label = ET.Element(
                'text',
                x=str(0),
                y=str(0),
                fill=theme.title_color
            )
            label.set('font-family', theme.title_font_family)
            label.set('font-size', str(theme.title_font_size))
            label.text = str(self._title)

            header_group.append(label)
            header_margin += theme.title_font_char_height + theme.title_gap

        # Legend
        if len(self._layers) > 1 or isinstance(self._layers[0][0], CategorySeries):
            legend_group = ET.Element('g')
            legend_group.set('transform', svg.translate(0, header_margin))

            indent = 0
            rows = 1
            palette = self._palette()

            for series, shape in self._layers:
                for item_group, item_width in shape.legend_to_svg(series, palette):
                    if indent + item_width > width:
                        indent = 0
                        rows += 1

                    y = (rows - 1) * (theme.legend_font_char_height + theme.legend_gap)
                    item_group.set('transform', svg.translate(indent, y))

                    indent += item_width

                    legend_group.append(item_group)

            legend_height = rows * (theme.legend_font_char_height + theme.legend_gap)

            header_margin += legend_height
            header_group.append(legend_group)

        margin_group.append(header_group)

        # Body
        body_group = ET.Element('g')
        body_group.set('transform', svg.translate(0, header_margin))

        body_width = margin_width
        body_height = margin_height - header_margin

        margin_group.append(body_group)

        # Axes
        x_scale, x_axis = self._validate_dimension(X)
        y_scale, y_axis = self._validate_dimension(Y)

        bottom_margin = x_axis.estimate_label_margin(x_scale, 'bottom')
        left_margin = y_axis.estimate_label_margin(y_scale, 'left')

        canvas_width = body_width - left_margin
        canvas_height = body_height - bottom_margin

        axes_group = ET.Element('g')
        axes_group.set('transform', svg.translate(left_margin, 0))

        axes_group.append(x_axis.to_svg(canvas_width, canvas_height, x_scale, 'bottom'))
        axes_group.append(y_axis.to_svg(canvas_width, canvas_height, y_scale, 'left'))

        header_group.set('transform', svg.translate(left_margin, 0))

        body_group.append(axes_group)

        # Series
        series_group = ET.Element('g')

        palette = self._palette()

        for series, shape in self._layers:
            series_group.append(shape.to_svg(canvas_width, canvas_height, x_scale, y_scale, series, palette))

        axes_group.append(series_group)

        return root_group

    def to_svg(self, path=None, width=None, height=None):
        """
        Render this chart to an SVG document.

        The :code:`width` and :code:`height` are specified in SVG's
        "unitless" units, however, it is usually convenient to specify them
        as though they were pixels.

        :param path:
            Filepath or file-like object to write to. If omitted then the SVG
            will be returned as a string. If running within IPython, then this
            will return a SVG object to be displayed.
        :param width:
            The output width, in SVG user units. Defaults to
            :data:`.theme.default_chart_width`.
        :param height:
            The output height, in SVG user units. Defaults to
            :data:`.theme.default_chart_height`.
        """
        width = width or theme.default_chart_width
        height = height or theme.default_chart_height

        root = ET.Element(
            'svg',
            width=str(width),
            height=str(height),
            version='1.1',
            xmlns='http://www.w3.org/2000/svg'
        )

        group = self.to_svg_group(width, height)
        root.append(group)

        svg_text = svg.stringify(root)
        close = True

        if path:
            f = None

            try:
                if hasattr(path, 'write'):
                    f = path
                    close = False
                else:
                    dirpath = os.path.dirname(path)

                    if dirpath and not os.path.exists(dirpath):
                        os.makedirs(dirpath)

                    f = open(path, 'w')

                f.write(svg.HEADER)
                f.write(svg_text)
            finally:
                if close and f is not None:
                    f.close()
        else:
            return IPythonSVG(svg_text)



"""
Helpers for working with SVG.
"""

import xml.etree.ElementTree as ET

HEADER = '<?xml version="1.0" standalone="no"?>\n' + \
    '<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"\n' + \
    '"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">\n'


def stringify(root):
    """
    Convert an SVG XML tree to a unicode string.
    """
    return ET.tostring(root, encoding='unicode')


def save(f, root):
    """
    Save an SVG XML tree to a file.
    """
    f.write(HEADER)
    f.write(stringify(root))


def translate(x, y):
    """
    Generate an SVG transform statement representing a simple translation.
    """
    return 'translate(%i %i)' % (x, y)


def rotate(deg, x, y):
    """
    Generate an SVG transform statement representing rotation around a given
    point.
    """
    return 'rotate(%i %i %i)' % (deg, x, y)



"""
This module contains all style configuration for rendering charts. Setting any
of these variables will change how charts are rendered.
"""

# CHART

#: Default chart width
default_chart_width = 800

#: Default chart height
default_chart_height = 600

#: Chart background color
background_color = '#f9f9f9'

#: Chart margin as a percent of chart width
margin = 0.05

# CHART TITLE

#: Chart title text color
title_color = '#333'

#: Chart title font
title_font_family = 'Monaco'

#: Chart title font size
title_font_size = 16

#: Approximate glyph height of the title font
title_font_char_height = 16

#: Approximate glyph width of the title font
title_font_char_width = 9

#: Gap between title and rest of chart
title_gap = 4

# LEGEND

#: Chart legend text color
legend_color = '#666'

#: Chart legend font
legend_font_family = 'Monaco'

#: Chart legend font size
legend_font_size = 14

#: Approximate glyph height of the legend font
legend_font_char_height = 14

#: Approximate glyph width of the legend font
legend_font_char_width = 8

#: Gap between legend and rest of chart
legend_gap = 4

#: Size of the bubble next to an legend item
legend_bubble_size = 10

#: Offset from the top of the glyph
legend_bubble_offset = 4

# AXIS

#: Axis title text color
axis_title_color = '#666'

#: Axis title font
axis_title_font_family = 'Monaco'

#: Axis title font size
axis_title_font_size = 14

#: Approximate glyph height of the axis title font
axis_title_font_char_height = 14

#: Approximate glyph width of the axis title font
axis_title_font_char_width = 8

#: Gap between axis title and rest of chart
axis_title_gap = 16

# TICKS

#: Width of a tick mark
tick_width = 1

#: Length of a tick mark
tick_size = 4

#: Color of tick marks
tick_color = '#eee'

#: Color of the zero tick mark
zero_color = '#a8a8a8'

# TICK LABELS

#: Color of tick label text
label_color = '#9c9c9c'

#: Tick label font
tick_font_family = 'Monaco'

#: Tick label font size
tick_font_size = 14

#: Approximate glyph height of the tick label font
tick_font_char_height = 14

#: Approximate glyph width of the tick label font
tick_font_char_width = 8

# SERIES

#: Default sequence of :class:`.Shape` colors
default_series_colors = ['#e41a1c', '#377eb8', '#4daf4a', '#984ea3', '#ff7f00']

#: Default :class:`.Dots` radius
default_dot_radius = 3

#: Default :class:`.Line` width
default_line_width = 2

#: Default stroke-dasharray property when using dashes on a line
default_stroke_dasharray = 'none'



from datetime import date, datetime
from decimal import Decimal


class DataType:
    """
    Base class for :class:`.Series` data types.
    """
    @classmethod
    def infer(cls, v):
        for t in [DateTime, Date, Number, Text]:
            if isinstance(v, t.types):
                return t

        raise TypeError('No data type available for %s' % type(v))


class Date(DataType):
    """
    Data representing dates.
    """
    types = (date,)


class DateTime(DataType):
    """
    Data representing dates with times.
    """
    types = (datetime,)


class Number(DataType):
    """
    Data representing numbers.
    """
    types = (int, float, Decimal)


class Text(DataType):
    """
    Data representing text/strings.
    """
    types = (str,)



import xml.etree.ElementTree as ET

from leather.data_types import Number, Text
from leather.series import CategorySeries
from leather.shapes.base import Shape
from leather.utils import X, Y


class Columns(Shape):
    """
    Render a series of data as columns.

    :param fill_color:
        The color to fill the columns. You may also specify a
        :func:`.style_function`.
    """
    def __init__(self, fill_color=None):
        self._fill_color = fill_color

    def validate_series(self, series):
        """
        Verify this shape can be used to render a given series.
        """
        if isinstance(series, CategorySeries):
            raise ValueError('Columns can not be used to render CategorySeries.')

        if series.data_type(X) is not Text:
            raise ValueError('Bars only support Text values for the X axis.')

        if series.data_type(Y) is not Number:
            raise ValueError('Bars only support Number values for the Y axis.')

    def to_svg(self, width, height, x_scale, y_scale, series, palette):
        """
        Render columns to SVG elements.
        """
        group = ET.Element('g')
        group.set('class', 'series columns')

        zero_y = y_scale.project(0, height, 0)

        if self._fill_color:
            fill_color = self._fill_color
        else:
            fill_color = next(palette)

        for d in series.data():
            if d.x is None or d.y is None:
                continue

            x1, x2 = x_scale.project_interval(d.x, 0, width)
            proj_y = y_scale.project(d.y, height, 0)

            if d.y < 0:
                column_y = zero_y
                column_height = proj_y - zero_y
            else:
                column_y = proj_y
                column_height = zero_y - proj_y

            if callable(fill_color):
                color = fill_color(d)
            else:
                color = fill_color

            group.append(ET.Element(
                'rect',
                x=str(x1),
                y=str(column_y),
                width=str(x2 - x1),
                height=str(column_height),
                fill=color
            ))

        return group



from collections import namedtuple
from datetime import date, datetime, timedelta
from decimal import Decimal

try:
    __IPYTHON__
    from IPython.display import SVG as IPythonSVG
except (NameError, ImportError):
    def IPythonSVG(x):
        return x


# Shorthand
ZERO = Decimal('0')
NINE_PLACES = Decimal('1e-9')

#: X data dimension index
X = 0

#: Y data dimension index
Y = 1

#: Z data dimension index
Z = 2


DIMENSION_NAMES = ['X', 'Y', 'Z']

#: Data structure for representing margins or other CSS-edge like properties
Box = namedtuple('Box', ['top', 'right', 'bottom', 'left'])

#: Data structure for a single series data point
Datum = namedtuple('Datum', ['i', 'x', 'y', 'z', 'row'])

#: Dummy object used in place of a series when rendering legends for categories
DummySeries = namedtuple('DummySeries', ['name'])


def to_year_count(d):
    """
    date > n years
    """
    return d.year


def from_year_count(n, t=date):
    """
    n years > date
    """
    return t(n, 1, 1)


def to_month_count(d):
    """
    date > n months
    """
    return (d.year * 12) + d.month


def from_month_count(n, t=date):
    """
    n months > date
    """
    return t(n // 12, (n % 12) + 1, 1)


def to_day_count(d):
    """
    date > n days
    """
    return (d - type(d).min).days


def from_day_count(n, t=date):
    """
    n days > date
    """
    return t.min + timedelta(days=n)


def to_hour_count(d):
    """
    date > n hours
    """
    return (d - datetime.min).total_seconds() / (60 * 60)


def from_hour_count(n, t=datetime):
    """
    n hours > date
    """
    return t.min + timedelta(hours=n)


def to_minute_count(d):
    """
    date > n minutes
    """
    return (d - datetime.min).total_seconds() / 60


def from_minute_count(n, t=datetime):
    """
    n minutes > date
    """
    return t.min + timedelta(minutes=n)


def to_second_count(d):
    """
    date > n seconds
    """
    return (d - datetime.min).total_seconds()


def from_second_count(n, t=datetime):
    """
    n seconds > date
    """
    return t.min + timedelta(seconds=n)


def to_microsecond_count(d):
    """
    date > n microseconds
    """
    return (d - datetime.min).total_seconds() * 1000


def from_microsecond_count(n, t=datetime):
    """
    n microseconds > date
    """
    return t.min + timedelta(microseconds=n)
